package ut.maxorato.bsc.crawler.analysis;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import ut.maxorato.bsc.crawler.generator.NewUrlGenerator;
import ut.maxorato.bsc.crawler.model.AnalysisContext;
import ut.maxorato.bsc.crawler.model.AnalysisReport;
import ut.maxorato.bsc.crawler.model.RequestContext;
import ut.maxorato.bsc.crawler.model.TemplateResult;
import ut.maxorato.bsc.crawler.model.UrlTemplate;

import java.io.File;
import java.io.FileWriter;
import java.io.IOException;

public class Analysis implements Runnable {
    private static final Logger log = LoggerFactory.getLogger(Analysis.class);

    private final File urlListFile;
    private final File resultLocation;

    public Analysis(File urlListFile, File resultLocation) {
        this.urlListFile = urlListFile;
        this.resultLocation = resultLocation;
    }

    private void writeReport(FileWriter resultWriter, AnalysisReport report) throws IOException {
        resultWriter.write("urlCount: " + report.urlCount + "\n");
        resultWriter.write("initialTemplateCount: " + report.initialTemplateCount + "\n");
        resultWriter.write("initialGroupCount: " + report.initialGroupCount + "\n");
        resultWriter.write("afterSplitTemplateCount: " + report.afterSplitTemplateCount + "\n");
        resultWriter.write("afterSplitGroupCount: " + report.afterSplitGroupCount + "\n");
        resultWriter.write("afterNumericFilterTemplateCount: " + report.afterNumericFilterTemplateCount + "\n");
        resultWriter.write("afterNumericFilterGroupCount: " + report.afterNumericFilterGroupCount + "\n");
        resultWriter.write("afterUnusedRemoverTemplateCount: " + report.afterUnusedRemoverTemplateCount + "\n");
        resultWriter.write("afterSingleNumericTemplateCount: " + report.afterSingleNumericTemplateCount + "\n");
        resultWriter.write("afterGetUnusedRemoverTemplateCount: " + report.afterGetUnusedRemoverTemplateCount + "\n");
        resultWriter.write("filteredTemplateCount: " + report.filteredTemplateCount + "\n\n");

        for (UrlTemplate template : report.allTemplates) {
            resultWriter.write(template.valueSetCount + " " + template.key + "\n");
        }

        resultWriter.write("\n");

        for (TemplateResult result : report.templateResults) {
            resultWriter.write(result.template.key + " " + result.totalAttempts + " " + result.generatedUrls.size() + " " + result.getRatio() + "\n");
        }

        resultWriter.write("\n");
        resultWriter.write(report.aggregateResults.totalAttempts + " " + report.aggregateResults.generatedUrls.size() + " " +  report.aggregateResults.getRatio() + "\n\n\n");

        for (TemplateResult result : report.templateResults) {
            resultWriter.write(result.template.key + " " + result.totalAttempts + " " + result.generatedUrls.size() + " " + result.getRatio() + "\n\n");
            resultWriter.write("FAILED:\n");

            for (String url : result.failedUrls) {
                resultWriter.write(url + "\n");
            }

            resultWriter.write("\nSUCCEEDED:\n");
            for (String url : result.generatedUrls) {
                resultWriter.write(url + "\n");
            }

            resultWriter.write("\n\n");
        }
    }

    @Override
    public void run() {
        RequestContext requestContext = new RequestContext();

        try (FileWriter resultWriter = new FileWriter(resultLocation)) {
            AnalysisContext analysisContext = new AnalysisContext(resultWriter);
            new UrlListReader(analysisContext, urlListFile).run();
            new InitialTemplateProducer(analysisContext).run();
            analysisContext.urls.clear();
            new TemplateGroupProducer(analysisContext).run();
            analysisContext.initialTemplates.clear();
            new MultiPartElementSplitter(analysisContext, requestContext).run();
            new GroupTemplateFilter(analysisContext, new MustHaveNumberParameterFilter()).run();
            new UnusedElementDetector(analysisContext, requestContext, false).run();
            new TemplateFinalizer(analysisContext, false).run();
            analysisContext.initialTemplates.clear();
            analysisContext.initialTemplates.putAll(analysisContext.finalTemplates);
            analysisContext.templateGroups.clear();
            analysisContext.finalTemplates.clear();
            new TemplateGroupProducer(analysisContext).run();
            analysisContext.initialTemplates.clear();
            new UnusedElementDetector(analysisContext, requestContext, true).run();
            new TemplateFinalizer(analysisContext, true).run();
            analysisContext.templateGroups.clear();
            new TemplateMapFilter(analysisContext.finalTemplates, new ValueSetSizeFilter()).run();
            analysisContext.report.filteredTemplateCount = analysisContext.finalTemplates.size();
            new NewUrlGenerator(analysisContext, requestContext).run();
            writeReport(resultWriter, analysisContext.report);
        } catch (Exception e) {
            log.error("Failed to process domain {}", urlListFile.getName(), e);
        }
    }
}
