package ut.maxorato.bsc.crawler.analysis;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import ut.maxorato.bsc.crawler.diff.RemovedFixComparisonStrategy;
import ut.maxorato.bsc.crawler.model.AnalysisContext;
import ut.maxorato.bsc.crawler.model.RequestContext;
import ut.maxorato.bsc.crawler.model.PathElement;
import ut.maxorato.bsc.crawler.model.PathElementType;
import ut.maxorato.bsc.crawler.model.UrlTemplate;
import ut.maxorato.bsc.crawler.model.UrlTemplateGroup;
import ut.maxorato.bsc.crawler.utils.StringUtils;

import java.util.ArrayList;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

public class MultiPartElementSplitter implements Runnable {
    private static final Logger log = LoggerFactory.getLogger(MultiPartElementSplitter.class);

    private final AnalysisContext context;
    private final RequestContext requestContext;
    private Map<String, UrlTemplateGroup> newGroups;

    public MultiPartElementSplitter(AnalysisContext context, RequestContext requestContext) {
        this.context = context;
        this.requestContext = requestContext;
        newGroups = new HashMap<>();
    }

    private String[] applySplitToText(SplitType type, String text) {
        switch (type) {
            case PREFIX:
                return StringUtils.getDigitPrefix(text);
            case SUFFIX:
                return StringUtils.getDigitSuffix(text);
            default:
                throw new IllegalArgumentException();
        }
    }

    private List<ByTextGrouping> getByTextGrouping(SplitMerge merge) {
        Map<String, ByTextGrouping> groupByText = new HashMap<>();

        for (TemplateWithSplitElement template : merge.templates) {
            ByTextGrouping grouping = groupByText.get(template.text);
            if (grouping == null) {
                grouping = new ByTextGrouping(template.text, merge.elementIndex, merge.splitType);
                groupByText.put(template.text, grouping);
            }
            grouping.templates.add(template);
        }

        List<ByTextGrouping> groupingList = new ArrayList<>(groupByText.values());

        groupingList.sort(new Comparator<ByTextGrouping>() {
            @Override
            public int compare(ByTextGrouping o1, ByTextGrouping o2) {
                return o2.templates.size() - o1.templates.size();
            }
        });

        return groupingList;
    }

    private boolean canMerge(RemovedFixComparisonStrategy strategy, ByTextGrouping groupOne, ByTextGrouping groupTwo) {
        return strategy.isValidAlternateUrl(groupOne.getFirstUrl(groupTwo.text, groupOne.first().number));
    }

    private boolean canMergeAll(List<ByTextGrouping> groupByText) {
        int verifyDepth = Math.min(3, groupByText.size());

        ByTextGrouping groupOne = groupByText.get(0);
        String originalPageUrl = groupOne.getFirstUrl(groupOne.text, groupOne.first().number);
        String emptyPageUrl = groupOne.getFirstUrl(groupOne.text, "999999999");
        RemovedFixComparisonStrategy strategy = new RemovedFixComparisonStrategy(requestContext, originalPageUrl, emptyPageUrl);

        for (int i = 1; i < verifyDepth; i++) {
            if (!canMerge(strategy, groupOne, groupByText.get(i))) {
                return false;
            }
        }

        return true;
    }

    private void removeFromGroup(UrlTemplateGroup group, List<TemplateWithSplitElement> templates) {
        for (TemplateWithSplitElement template : templates) {
            group.templates.remove(template.template.key);
        }
    }

    private List<PathElement> createSplitTemplate(SplitType splitType, int elementIndex, String staticText, TemplateWithSplitElement template) {
        List<PathElement> pathElements = new ArrayList<>();
        UrlTemplate base = template.template;

        for (int i = 0; i < base.elements.size(); i++) {
            if (i != elementIndex) {
                pathElements.add(base.elements.get(i));
            } else {
                PathElement number = new PathElement(PathElementType.NUM_VALUE, false, template.number);
                PathElement text = staticText == null ?
                        new PathElement(PathElementType.TEXT_VALUE, false, template.text) :
                        new PathElement(PathElementType.TEXT_STATIC, false, staticText);

                for (int j = 0; j < base.valueSetCount - 1; j++) {
                    number.values.add(template.number);
                    if (staticText == null) {
                        text.values.add(template.text);
                    }
                }

                if (splitType == SplitType.PREFIX) {
                    pathElements.add(number);
                    pathElements.add(text);
                } else {
                    pathElements.add(text);
                    pathElements.add(number);
                }
            }
        }

        return pathElements;
    }

    private void createNewTemplateBySplit(SplitType splitType, int elementIndex, String staticText, List<TemplateWithSplitElement> templates) {
        UrlTemplate template = new UrlTemplate(createSplitTemplate(splitType, elementIndex, staticText, templates.get(0)));
        template.valueSetCount = templates.get(0).template.valueSetCount;

        for (int i = 1; i < templates.size(); i++) {
            TemplateWithSplitElement mergedTemplate = templates.get(i);
            template.merge(createSplitTemplate(splitType, elementIndex, staticText, mergedTemplate));
            template.valueSetCount += mergedTemplate.template.valueSetCount;
        }

        String groupKey = template.getGroupKey();
        UrlTemplateGroup group = newGroups.get(groupKey);
        if (group == null) {
            group = new UrlTemplateGroup(groupKey);
            newGroups.put(groupKey, group);
        }

        group.templates.put(template.key, template);
    }

    private void applyMergeToAll(UrlTemplateGroup group, SplitMerge merge) {
        createNewTemplateBySplit(merge.splitType, merge.elementIndex, null, merge.templates);
        removeFromGroup(group, merge.templates);
    }

    private void applyMergePerTextGroup(UrlTemplateGroup group, SplitMerge merge, List<ByTextGrouping> groupByText) {
        for (ByTextGrouping withSameText : groupByText) {
            if (withSameText.templates.size() < 10) {
                continue;
            }

            createNewTemplateBySplit(merge.splitType, merge.elementIndex, withSameText.text, withSameText.templates);
            removeFromGroup(group, withSameText.templates);
        }
    }

    private void verifyAndApplyMerge(UrlTemplateGroup group, SplitMerge merge) {
        if (merge.numbers.size() < merge.templates.size()) {
            return;
        } else if (merge.numbers.size() < 10) {
            return;
        }

        log.debug("Verifying prefix/suffix merge key {}", merge.mergeKey);

        List<ByTextGrouping> groupByText = getByTextGrouping(merge);

        if (canMergeAll(groupByText)) {
            applyMergeToAll(group, merge);
        } else {
            applyMergePerTextGroup(group, merge, groupByText);
        }
    }

    private void processGroupElement(UrlTemplateGroup group, int index) {
        SplitForGroup groupSplit = new SplitForGroup(index);

        for (UrlTemplate template : group.templates.values()) {
            PathElement element = template.elements.get(index);

            if (element.type != PathElementType.TEXT_STATIC || element.isGetPart) {
                return;
            }

            for (SplitType type : SplitType.values()) {
                String[] parts = applySplitToText(type, element.values.get(0));

                if (parts != null) {
                    String mergeKey = template.getKeyWithConstant(index, "{N+T}");
                    SplitMerge[] merges = groupSplit.fetch(mergeKey);

                    merges[type.ordinal()].numbers.add(Integer.parseInt(parts[0]));
                    merges[type.ordinal()].templates.add(new TemplateWithSplitElement(template, parts[0], parts[1]));
                }
            }
        }

        for (SplitMerge[] merges : groupSplit.splitMerges.values()) {
            SplitMerge merge = merges[0].numbers.size() >= merges[1].numbers.size() ? merges[0] : merges[1];
            verifyAndApplyMerge(group, merge);
        }
    }

    private void processGroup(UrlTemplateGroup group) {
        int elementCount = group.getPathElementCount();

        for (int i = elementCount - 1; i > 0; i--) {
            processGroupElement(group, i);
        }
    }

    @Override
    public void run() {
        int finalTemplateCount = 0;

        for (Iterator<UrlTemplateGroup> iterator = context.templateGroups.values().iterator(); iterator.hasNext();) {
            UrlTemplateGroup group = iterator.next();
            processGroup(group);
            finalTemplateCount += group.templates.size();

            if (group.templates.size() == 0) {
                iterator.remove();
            }
        }

        while (newGroups.size() > 0) {
            List<UrlTemplateGroup> iterationGroups = new ArrayList<>(newGroups.values());
            newGroups.clear();

            for (UrlTemplateGroup group : iterationGroups) {
                processGroup(group);

                if (group.templates.size() > 0) {
                    context.templateGroups.put(group.key, group);
                    finalTemplateCount += group.templates.size();
                }
            }
        }

        context.report.afterSplitTemplateCount = finalTemplateCount;
        context.report.afterSplitGroupCount = context.templateGroups.size();

        log.info("Group count changed to {}.", context.templateGroups.size());
    }

    private enum SplitType {
        PREFIX,
        SUFFIX
    }

    private static class ByTextGrouping {
        public final int elementIndex;
        public final SplitType splitType;
        public final String text;
        public final List<TemplateWithSplitElement> templates;

        private ByTextGrouping(String text, int elementIndex, SplitType splitType) {
            this.text = text;
            this.elementIndex = elementIndex;
            this.splitType = splitType;
            templates = new ArrayList<>();
        }

        private TemplateWithSplitElement first() {
            return templates.get(0);
        }

        private String getFirstUrl(String text, String number) {
            String replacement = splitType == SplitType.PREFIX ? number + text : text + number;
            return templates.get(0).template.nthUrlWithPlaceholder(0, elementIndex).replace("{Y}", replacement);
        }
    }

    private static class SplitMerge {
        public final int elementIndex;
        public final String mergeKey;
        public final SplitType splitType;
        public final Set<Integer> numbers;
        public final List<TemplateWithSplitElement> templates;

        public SplitMerge(int elementIndex, String mergeKey, SplitType splitType) {
            this.elementIndex = elementIndex;
            this.mergeKey = mergeKey;
            this.splitType = splitType;
            this.numbers = new HashSet<>();
            this.templates = new ArrayList<>();
        }
    }

    private static class SplitForGroup {
        public final int elementIndex;
        public final Map<String, SplitMerge[]> splitMerges;

        public SplitForGroup(int elementIndex) {
            this.elementIndex = elementIndex;
            this.splitMerges = new HashMap<>();
        }

        public SplitMerge[] fetch(String key) {
            SplitMerge[] splitMerge = splitMerges.get(key);

            if (splitMerge == null) {
                splitMerge = new SplitMerge[] {
                        new SplitMerge(elementIndex, key, SplitType.PREFIX),
                        new SplitMerge(elementIndex, key, SplitType.SUFFIX)
                };

                splitMerges.put(key, splitMerge);
            }

            return splitMerge;
        }
    }

    private static class TemplateWithSplitElement {
        public final UrlTemplate template;
        public final String number;
        public final String text;

        private TemplateWithSplitElement(UrlTemplate template, String number, String text) {
            this.template = template;
            this.number = number;
            this.text = text;
        }
    }
}
