package ut.maxorato.bsc.crawler.analysis;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import ut.maxorato.bsc.crawler.diff.RemovedCategoryComparisonStrategy;
import ut.maxorato.bsc.crawler.diff.RemovedNameComparisonStrategy;
import ut.maxorato.bsc.crawler.model.AnalysisContext;
import ut.maxorato.bsc.crawler.model.RequestContext;
import ut.maxorato.bsc.crawler.model.PathElement;
import ut.maxorato.bsc.crawler.model.PathElementType;
import ut.maxorato.bsc.crawler.model.UrlTemplate;
import ut.maxorato.bsc.crawler.model.UrlTemplateGroup;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

public class UnusedElementDetector implements Runnable {
    private static final Logger log = LoggerFactory.getLogger(UnusedElementDetector.class);

    private final AnalysisContext context;
    private final RequestContext requestContext;
    private final boolean doGetParameters;

    public UnusedElementDetector(AnalysisContext context, RequestContext requestContext, boolean doGetParameters) {
        this.context = context;
        this.requestContext = requestContext;
        this.doGetParameters = doGetParameters;
    }

    private float findMainVariableUniqueValueRatio(List<UrlTemplate> templates) {
        UrlTemplate first = templates.get(0);
        float bestRatio = 0.0f;

        for (int i = 0; i < first.elements.size(); i++) {
            if (first.elements.get(i).type == PathElementType.NUM_VALUE) {
                Set<String> uniqueValues = new HashSet<>();
                int totalUrlCount = 0;

                for (UrlTemplate template : templates) {
                    uniqueValues.addAll(template.elements.get(i).values);
                    totalUrlCount += template.valueSetCount;
                }

                if (totalUrlCount > 10) {
                    bestRatio = Math.max(bestRatio, (float) uniqueValues.size() / (float) totalUrlCount);
                }
            }
        }

        return bestRatio;
    }

    private boolean canNameMerge(RemovedNameComparisonStrategy strategy, String otherUrl) {
        return strategy.isValidAlternateUrl(otherUrl);
    }

    private boolean verifyNameMerge(Merge merge) {
        int verifyDepth = Math.min(3, merge.templates.size());

        UrlTemplate templateOne = merge.templates.get(0);
        String keyWithPlaceholder = templateOne.nthUrlWithPlaceholder(0, merge.elementIndex);
        String originalValue = templateOne.getFirstValue(merge.elementIndex);
        String originalPageUrl = keyWithPlaceholder.replace("{Y}", originalValue);
        RemovedNameComparisonStrategy strategy = new RemovedNameComparisonStrategy(requestContext, originalPageUrl);

        for (int i = 1; i < verifyDepth; i++) {
            String value = merge.templates.get(i).getFirstValue(merge.elementIndex);
            String alternateUrl = keyWithPlaceholder.replace("{Y}", value);

            if (!canNameMerge(strategy, alternateUrl)) {
                return false;
            }
        }

        return true;
    }

    private boolean canCategoryMerge(RemovedCategoryComparisonStrategy strategy, String otherUrl) {
        return strategy.isValidAlternateUrl(otherUrl);
    }

    private boolean verifyCategoryMerge(Merge merge) {
        Collections.sort(merge.templates, new Comparator<UrlTemplate>() {
            @Override
            public int compare(UrlTemplate o1, UrlTemplate o2) {
                return o2.valueSetCount - o1.valueSetCount;
            }
        });

        int verifyDepth = Math.min(3, merge.templates.size());

        UrlTemplate templateOne = merge.templates.get(0);
        String originalText = templateOne.getFirstValue(merge.elementIndex);
        String originalOneKey = templateOne.nthUrlWithPlaceholder(0, merge.elementIndex);
        String originalTwoKey = templateOne.nthUrlWithPlaceholder(1, merge.elementIndex);
        String originalOneUrl = originalOneKey.replace("{Y}", originalText);
        String originalTwoUrl = originalTwoKey.replace("{Y}", originalText);

        RemovedCategoryComparisonStrategy strategy = new RemovedCategoryComparisonStrategy(requestContext, originalOneUrl, originalTwoUrl);

        for (int i = 1; i < verifyDepth; i++) {
            String value = merge.templates.get(i).getFirstValue(merge.elementIndex);
            String alternateUrl = originalOneKey.replace("{Y}", value);

            if (!canCategoryMerge(strategy, alternateUrl)) {
                return false;
            }
        }

        return true;
    }

    private void applyMerge(UrlTemplateGroup group, Merge merge) {
        UrlTemplate first = merge.first();
        List<PathElement> newElements = first.elements;
        List<String> values = new ArrayList<>(Collections.nCopies(first.valueSetCount, first.getFirstValue(merge.elementIndex)));
        PathElement changedElement = new PathElement(PathElementType.TEXT_VALUE, false, values);
        newElements.set(merge.elementIndex, changedElement);
        UrlTemplate newTemplate = new UrlTemplate(newElements, first.valueSetCount);

        group.templates.remove(first.key);

        for (int i = 1; i < merge.templates.size(); i++) {
            UrlTemplate template = merge.templates.get(i);
            newTemplate.merge(template);

            if (template.valueSetCount > 1) {
                List<String> newValues = Collections.nCopies(template.valueSetCount - 1, template.getFirstValue(merge.elementIndex));
                changedElement.values.addAll(newValues);
            }

            group.templates.remove(template.key);
        }

        group.templates.put(newTemplate.key, newTemplate);
    }

    private void verifyAndApplyMerge(UrlTemplateGroup group, Merge merge) {
        if (merge.templates.size() < 2 || findMainVariableUniqueValueRatio(merge.templates) < 0.95f) {
            return;
        }

        int totalUrlCount = 0;
        for (UrlTemplate template : merge.templates) {
            totalUrlCount += template.valueSetCount;
        }

        log.debug("Verifying unused element merge key {}", merge.mergeKey);

        boolean result = false;

        if (merge.templates.size() >= 0.95 * totalUrlCount) {
            if (merge.templates.size() > 10 && !merge.isGetPart) {
                result = verifyNameMerge(merge);
            }
        } else {
            result = verifyCategoryMerge(merge);
        }

        if (result) {
            applyMerge(group, merge);
        }
    }

    private void processGroupElement(UrlTemplateGroup group, int index) {
        MergeForGroup groupMerge = new MergeForGroup(index);

        for (UrlTemplate template : group.templates.values()) {
            PathElement element = template.elements.get(index);

            if (element.type != PathElementType.TEXT_STATIC || element.isGetPart != doGetParameters) {
                return;
            }

            String mergeKey = template.getKeyWithConstant(index, "{Y}");
            Merge merge = groupMerge.fetch(mergeKey, element.isGetPart);

            merge.templates.add(template);
        }

        for (Merge merge : groupMerge.merges.values()) {
            verifyAndApplyMerge(group, merge);
        }
    }

    public void processGroup(UrlTemplateGroup group) {
        int elementCount = group.getPathElementCount();

        for (int i = elementCount - 1; i > 0; i--) {
            processGroupElement(group, i);
        }
    }

    @Override
    public void run() {
        for (UrlTemplateGroup group : context.templateGroups.values()) {
            processGroup(group);
        }

        int finalTemplateCount = 0;
        for (UrlTemplateGroup group : context.templateGroups.values()) {
            finalTemplateCount += group.templates.size();
        }

        if (doGetParameters) {
            context.report.afterGetUnusedRemoverTemplateCount = finalTemplateCount;
        } else {
            context.report.afterUnusedRemoverTemplateCount = finalTemplateCount;
        }
    }

    private static class MergeForGroup {
        public final int elementIndex;
        public final Map<String, Merge> merges;

        public MergeForGroup(int elementIndex) {
            this.elementIndex = elementIndex;
            this.merges = new HashMap<>();
        }

        public Merge fetch(String key, boolean isGetPart) {
            Merge merge = merges.get(key);

            if (merge == null) {
                merge = new Merge(elementIndex, key, isGetPart);
                merges.put(key, merge);
            }

            return merge;
        }
    }

    private static class Merge {
        public final int elementIndex;
        public final String mergeKey;
        public final List<UrlTemplate> templates;
        public final boolean isGetPart;

        public Merge(int elementIndex, String mergeKey, boolean isGetPart) {
            this.elementIndex = elementIndex;
            this.mergeKey = mergeKey;
            this.templates = new ArrayList<>();
            this.isGetPart = isGetPart;
        }

        public UrlTemplate first() {
            return templates.get(0);
        }
    }
}
