package ut.maxorato.bsc.crawler.split;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.Writer;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

public class LogSplitter {
    private static final Logger log = LoggerFactory.getLogger(LogSplitter.class);

    private final static int MEMORY_LINE_LIMIT = 10000;

    private final Map<String, SiteData> siteDataMap;
    private final File outputDirectory;
    private final List<String> schemes;

    private int linesStored;

    public LogSplitter(File outputDirectory) {
        this.outputDirectory = outputDirectory;

        siteDataMap = new HashMap<>();
        schemes = new ArrayList<>();
        schemes.add("http");
        schemes.add("https");
        linesStored = 0;
    }

    private void flushSiteData(SiteData siteData) throws IOException {
        File outputFile = new File(outputDirectory, siteData.domain + ".txt");
        Writer writer = new BufferedWriter(new FileWriter(outputFile, true));

        for (String line : siteData.paths) {
            writer.write(line + "\n");
        }

        writer.close();

        linesStored -= siteData.paths.size();
        siteData.paths.clear();
    }

    private void flush(int threshold) throws IOException {
        for (SiteData siteData : siteDataMap.values()) {
            if (siteDataMap.size() >= threshold) {
                flushSiteData(siteData);
            }
        }
    }

    private void partialFlush() throws IOException {
        flush(linesStored / siteDataMap.size());
    }

    private void storeEntry(String domain, String url) throws IOException {
        SiteData siteData = siteDataMap.get(domain);

        if (siteData == null) {
            siteData = new SiteData(domain);
            siteDataMap.put(domain, siteData);
        }

        siteData.paths.add(url);

        if (linesStored >= MEMORY_LINE_LIMIT) {
            partialFlush();
        }
    }

    private void processLogUrl(String url) throws IOException {
        int slashPosition = url.indexOf('/', url.indexOf("://") + 3);

        if (slashPosition == -1) {
            log.warn("Ignoring url with no slash: {}", url);
        } else {
            String domain = url.substring(0, slashPosition).replaceAll("[^a-zA-Z0-9]", "_");

            storeEntry(domain, url);
        }
    }

    private void processLogLine(String line) throws IOException {
        String[] parts = line.split(" +");

        if (parts.length < 12) {
            //log.warn("Ignoring line with too few parts: {}", line);
        } else if (!"200".equals(parts[1])) {
            //log.warn("Ignoring line with non-200 code: {}", line);
        } else if (!"text/html".equals(parts[6])) {
            //log.warn("Ignoring line with non-text/html type: {}", line);
        } else {
            String url = parts[3];

            int schemePosition = -1;

            for (String scheme : schemes) {
                if (url.startsWith(scheme + "://")) {
                    schemePosition = scheme.length();
                    break;
                }
            }

            if (schemePosition == -1) {
                log.warn("Ignoring line with unsupported scheme: {}", url);
            } else {
                processLogUrl(url);
            }
        }
    }

    public void splitLogFile(File logFile) throws IOException {
        if (!outputDirectory.exists() && !outputDirectory.mkdirs()) {
            log.error("Could not create output directory.");
            return;
        }

        BufferedReader reader = new BufferedReader(new FileReader(logFile));
        String line;
        long totalProcessed = 0;
        long lastReported = 0;

        while((line = reader.readLine()) != null) {
            processLogLine(line);
            totalProcessed += line.length() + 1;

            if (totalProcessed - lastReported > 50000000) {
                log.info("Processed {} bytes from {}", totalProcessed, logFile.getName());
                lastReported = totalProcessed;
            }
        }

        log.info("Finished processing {} bytes from {}", totalProcessed, logFile.getName());

        flush(0);

        reader.close();
    }
}
